//
// Dual port memory
// Two read/write access ports
//
//
// Freeware September 2015, Fen Logic Ltd.
// This code is free and is delivered 'as is'.
// It comes without any warranty, to the extent permitted by applicable law.
// There are no restrictions to any use or re-use of this code
// in any form or shape. It would be nice if you keep my company name
// in the source or modified source code but even that is not
// required as I can't check it anyway.
// But the code comes with no warranties or guarantees whatsoever.
//
//    This code maps the logic onto a XILINX embedded memory
//
// Known issues:
//    None
//
//
//  Port clash checking
//  ===================
//  You can not normally access a memory location that is being written
//  from the other port (read or write), near simultaneously.
//  The program has a built-in check for that (CHECK=1)
//  The write happens on the rising edge of the clock (either port)
//  The user can specify a 'forbidden time' before and after the clock edge
//  where the other side is not allowed to read or write the location
//  which has just been written
//  PRETIME  is the time before the clock edge
//  POSTTIME is the time after  the clock edge
//            +--------------+             +--------------+     
//    CLK     |              |             |              |     
//   ---------+              +-------------+              +-----
//                                                     
//            +----------------------------+           
//   WRITE    |                            |           
//  ----------+                            +--------------------
//                               PRETIME      POSTTIME
//                              <----------|------------>
//                              +-----------------------+
//                              |      FORBIDDEN        |
//                              +-----------------------+
//                                                           
//  Footnote: you can do this with $setuphold and negative timing but 
//  negative timing is non trivial thus I avoid that here.

module dpm
#(parameter WIDTH    = 8, // width in bits
            L2D      = 8, // Log 2 Depth, 8 = 256 deep 
            // Simulation parameters
            CLEAR    = 0,
            FILE     = "",
            // Access clash check
            CHECK    = 0,   // 1 = on, 0 = off
            PRETIME  = 10,  // Time before edge
            POSTTIME = 20   // Time after edge
)
(
   input                  a_clk,     // clock  
   input                  a_enable,  // enable               
   input                  a_write,   // write access 
   input        [L2D-1:0] a_adrs,    // address
   input      [WIDTH-1:0] a_wdata,   // write data
   output reg [WIDTH-1:0] a_rdata,   // read data
   
   input                  b_clk,     // clock  
   input                  b_enable,  // enable               
   input                  b_write,   // write access 
   input        [L2D-1:0] b_adrs,    // address
   input      [WIDTH-1:0] b_wdata,   // write data
   output reg [WIDTH-1:0] b_rdata    // read data
);

localparam DEPTH = (1<<L2D);

reg [WIDTH-1:0] memory [0:DEPTH-1];

   always @(posedge a_clk)
   begin
      if (a_enable)
      begin
         if (a_write)
            memory[a_adrs] <= a_wdata;
         else
            a_rdata <= memory[a_adrs];
      end
   end
         
   always @(posedge b_clk)
   begin
      if (b_enable)
      begin
         if (b_write)
            memory[b_adrs] <= b_wdata;
         else
            b_rdata <= memory[b_adrs];
      end
   end
   
/*
  synthesis translate_...
  synopsys  translate_...
  pragma    translate_...
 */
//synopsys translate_off

  
generate  
if (CHECK)
begin
reg   a_critical,b_critical;
event a_check,b_check;
reg  [L2D-1:0] a_check_adrs,a_access_adrs;    
reg  [L2D-1:0] b_check_adrs,b_access_adrs;    


   always @(posedge a_clk)
   begin
      // on write start critical A time
      if (a_enable & a_write)
      begin
         a_check_adrs <= a_adrs;
         a_critical <= 1'b1;
         a_critical <= #(PRETIME+POSTTIME) 1'b0;
      end
      
      // If enable (read or write) check critical other side
      if (a_enable)
      begin
         a_access_adrs <= a_adrs;
         #PRETIME -> b_check;      
      end
      
   end
         
   always @(posedge b_clk)
   begin
      // on write start critical B time
      if (b_enable & b_write)
      begin
         b_check_adrs <= b_adrs;
         b_critical <= 1'b1;
         b_critical <= #(PRETIME+POSTTIME) 1'b0;
      end
      // If enable (read or write) check critical other side
      if (b_enable)
      begin
         b_access_adrs <= b_adrs;
         #PRETIME -> a_check;      
      end
   end
   
   always @(a_check)
      if (a_critical===1'b1 && b_access_adrs==a_check_adrs)
      begin
         $display("%m@%0t B access clash with A write",$time);
         #1 $stop;
      end

   always @(b_check)
      if (b_critical===1'b1 && a_access_adrs==b_check_adrs)
      begin
         $display("%m@%0t A access clash with B write",$time);
         #1 $stop;
      end

end 
endgenerate

integer i;
   initial
   begin
      if (CLEAR)
      begin
         for (i=0; i<DEPTH; i=i+1)
            memory[i] = 0;
      end
      #0; // prevent simultaneous clear and file loading 
      // (Yes: it is all blocking code but still I have seen it happen)
      
      // If the filename is not zero use it to load from
      if (FILE[7:0]!=0)
         $readmemh(FILE,memory);
   end
//synopsys translate_on

endmodule