//
// Two port memory
// One write and and one read port
//
// Freeware September 2015, Fen Logic Ltd.
// This code is free and is delivered 'as is'.
// It comes without any warranty, to the extent permitted by applicable law.
// There are no restrictions to any use or re-use of this code
// in any form or shape. It would be nice if you keep my company name
// in the source or modified source code but even that is not
// required as I can't check it anyway.
// But the code comes with no warranties or guarantees whatsoever.
//
//    This code maps the logic onto a XILINX embedded memory
//
// Known issues:
//    None
//
// Two port memorys have one read and one write port in contrast to 
// dual-ported memory which have two read and two write ports.
// Two port memories can be made smaller then dual-ported memories
// and are so common that they deserve to be destinguished 
// from dual-ported memories (At least in the ASIC world!)
//
//
//  read/write clash checking
//  ===================
//  You can not normally read a memory location that is being written
//  from the other port, near simultaneously.
//  The program has a built-in check for that (CHECK=1)
//  The write happens on the rising edge of the clock.
//  The user can specify a 'forbidden time' before and after the clock edge
//  where the other side is not allowed to read the location
//  which has just been written
//  PRETIME  is the time before the clock edge
//  POSTTIME is the time after  the clock edge
//            +--------------+             +--------------+     
//    CLK     |              |             |              |     
//   ---------+              +-------------+              +-----
//                                                     
//            +----------------------------+           
//   WRITE    |                            |           
//  ----------+                            +--------------------
//                               PRETIME      POSTTIME
//                              <----------|------------>
//                              +-----------------------+
//                              |      FORBIDDEN        |
//                              +-----------------------+
//                                                           
//  Footnote: you can do this with $setuphold and negative timing but 
//  negative timing is non trivial thus I avoid that here.

module tpm
#(parameter WIDTH    = 8, // width in bits
            L2DEPTH  = 8, // Log 2 Depth, 8 = 256 deep 
            // Simulation parameters
            CLEAR    = 0,
            FILE     = "",
            // read/write clash check
            CHECK    = 0,   // 1 = on, 0 = off
            PRETIME  = 10,  // Time before edge
            POSTTIME = 20   // Time after edge
)
(
   input                  wclk,    // write clock                
   input                  write,   // write access 
   input    [L2DEPTH-1:0] wadrs,   // write address
   input      [WIDTH-1:0] wdata,   // write data
   
   input                  rclk,    // read clock                
   input                  read,    // read access
   input    [L2DEPTH-1:0] radrs,   // read address
   output reg [WIDTH-1:0] rdata    // read data
);

localparam DEPTH = (1<<L2DEPTH);

reg [WIDTH-1:0] memory [0:DEPTH-1];

   always @(posedge wclk)
      if (write)
         memory[wadrs] <= wdata;
         
   always @(posedge rclk)
      if (read)
         rdata <= memory[radrs];
   
   
/*
  synthesis translate_...
  synopsys  translate_...
  pragma    translate_...
 */
//synopsys translate_off

//
// Direct read and write access operating in
// zero time for testbenches
//

function [WIDTH-1:0] do_read;
input [L2DEPTH-1:0] address;
begin
   do_read = memory[address];
end
endfunction

task do_write;
input [L2DEPTH-1:0] address;
input [WIDTH-1:0] data;
begin
   memory[address] = data;
end
endtask

  
generate  
if (CHECK)
begin
reg   critical;
event check;
reg  [L2DEPTH-1:0] check_adrs,read_adrs;    
   

   always @(posedge wclk)
   begin
      // on write start critical time
      if (write)
      begin
         check_adrs <= wadrs;
         critical <= 1'b1;
         critical <= #(PRETIME+POSTTIME) 1'b0;
      end
   end
         
   always @(posedge rclk)
   begin
      // If read check critical other side
      if (read)
      begin
         read_adrs <= radrs;
         #PRETIME -> check;      
      end
    end
   
   always @(check)
      if (critical===1'b1 && read_adrs==check_adrs)
      begin
         $display("%m@%0t read/write clash",$time);
         #1 $stop;
      end

end 
endgenerate

integer i;
   initial
   begin
      if (CLEAR)
      begin
         for (i=0; i<DEPTH; i=i+1)
            memory[i] = 0;
      end
      #0; // prevent simultaneous clear and file loading 
      // (Yes: it is all blocking code but still I have seen it happen)
      
      // If the filename is not zero use it to load from
      if (FILE[7:0]!=0)
         $readmemh(FILE,memory);
   end
//synopsys translate_on

endmodule